import copy
import io
import logging
import traceback

from avro import io as avro_io
from avro import ipc as avro_ipc
from avro import protocol as avro_protocol
from avro import schema as avro_schema
from pyramid import response as p_response
from pyramid import threadlocal as p_threadlocal
from webob import exc as http_exc
from zope import interface as zi

logger = logging.getLogger(__name__)


class ServiceResponder(avro_ipc.Responder):
    """
    An Avro service responder which executes a callback to get a response.
    """

    def __init__(self, executor, *args, **kwargs):
        """
        Overridden init to add an executor callback.

        :param executor: a callback to use for retrieving a response.
        :param args: regular avro.ipc.Responder args.
        :param kwargs: regular avro.ipc.Responder kwargs.
        """
        self.executor = executor
        super(ServiceResponder, self).__init__(*args, **kwargs)

    def invoke(self, msg, req):
        """
        Call self.executor, then verify that the response fits the protocol
        that this knows how to speak.

        :param msg: an avro message.
        :param req: request arguments.
        :return: an avro response.
        """
        response = self.executor(msg.name, **req)
        local_message = self.local_protocol.messages.get(msg.name)
        local_response = local_message.response
        is_valid = avro_io.validate(local_response, response)
        if not is_valid:
            err = "Server response did not conform to its local schema."
            logging.error(
                "{}; Response: {}, Schema: {}".format(
                    err, response, local_response
                )
            )
            raise avro_ipc.AvroRemoteException(err)
        return response


class IAvroServiceRoute(zi.Interface):

    protocol = zi.Attribute("""Avro protocol object.""")
    responder = zi.Attribute("""Avro service responder.""")


@zi.implementer(IAvroServiceRoute)
class AvroServiceRoute(object):

    protocol = None
    responder = None

    def __init__(self, path, schema):
        self.path = path
        self.dispatch = {}
        self.protocol = avro_protocol.parse(schema)
        self.responder = ServiceResponder(self.execute_command, self.protocol)

    def register_message_impl(self, message, message_impl):
        if self.protocol.messages.get(message) is None:
            raise avro_schema.AvroException(
                "Message '{}' not defined.".format(message)
            )

        self.dispatch[message] = message_impl

    def validate_request(self, request):
        """
        Place for validating an incoming request.

        Right now, it simply verifies that there is a non-zero content-length.

        Any additional validation required will be added here in the future.

        :param request: a webob request.
        """
        content_length = int(request.headers["content-length"])
        if request.body_file is None or content_length == 0:
            raise http_exc.HTTPBadRequest()

    def __call__(self, request):
        """
        Reads the avro request data, then call our responder to respond.

        This will end up executing "execute_command" below.

        After getting a response from the responder, form a pyramid response
        and return it.

        This is basically the view object registered as a pyramid route.

        TODO: Figure out a way to send the request object through to the
        handler such that an implementation can modify any attributes of that
        request that it needs.

        :param request: a pyramid request.
        :return: a pyramid response.
        """
        self.validate_request(request)
        reader = avro_ipc.FramedReader(request.body_file)
        try:
            request_data = reader.read_framed_message()
        except avro_ipc.ConnectionClosedException:
            logger.exception("Failed to process request.")
            return http_exc.HTTPBadRequest()

        try:
            rpc_response = self.responder.respond(request_data)
        except http_exc.HTTPException as ex:
            logger.exception("HTTP exception while processing message.")
            return ex
        except Exception:
            logger.exception("Error processing RPC content.")
            return http_exc.HTTPInternalServerError()

        with io.BytesIO() as _body_file:
            writer = avro_ipc.FramedWriter(_body_file)
            writer.write_framed_message(rpc_response)
            response_body = _body_file.getvalue()

        logger.debug("Finished request. Returning response.")
        return p_response.Response(
            status=200,
            body=response_body,
            headerlist=[("Content-Type", "avro/binary")]
        )

    def execute_command(self, command, **command_args):
        """
        Given the provided command and its arguments, retrieve a registered
        message callback and execute it.

        Prior to executing the registered callback, attach the provided
        arguments as an "avro_data" attribute on the request object.

        :param command: an avro message name.
        :param command_args: avro message arguments.
        :return: a response from the handler.
        """
        exists = command in self.protocol.props["messages"].keys()
        if not exists:
            err = "Message not found: '{}'".format(command)
            raise avro_ipc.AvroRemoteException(err)

        handler = self.dispatch.get(command)
        if handler is None:
            err = "No handler registered for: '{}'".format(command)
            raise avro_ipc.AvroRemoteException(err)

        try:
            logger.debug("Invoking handler {}".format(handler))
            request = p_threadlocal.get_current_request()
            request.avro_data = copy.deepcopy(command_args)
            response = handler(request)
        except Exception:
            logging.exception("Error handling request: {}".format(command))
            raise avro_ipc.AvroRemoteException(traceback.format_exc())

        return response


__all__ = [IAvroServiceRoute.__name__, AvroServiceRoute.__name__]



import logging

from pyramid import config as p_config
from pyramid import settings as p_settings

logger = logging.getLogger(__name__)

CONFIG_PREFIX = "avro."

CONFIG_DEFAULTS = {
    "default_path_prefix": None,
    "protocol_dir": None,
    "auto_compile": False,
    "tools_jar": None,
    "service": {}
}

SERVICE_DEF_PROPERTIES = frozenset((
    "protocol",
    "schema",
    "pattern"
))


def get_config_options(configuration):
    """
    Given a dictionary of config options, normalize and produce a resultant
    config dict for any keys prepended with "avro."

    The expectation here is that the configuration dict came straight from
    the result of parsing a config (.ini) file.

    :param configuration: a config dict.
    :return: a dict of config values for this plugin.
    """
    options = CONFIG_DEFAULTS.copy()
    parsed_options = dict(
        (key[len(CONFIG_PREFIX):], val)
        for key, val in configuration.items()
        if key.startswith(CONFIG_PREFIX)
    )
    for key in parsed_options.keys():
        val = parsed_options[key]
        if key.startswith("service."):
            if not isinstance(val, basestring) or not val:
                raise p_config.ConfigurationError(
                    "Service definition must have one on {}".format(
                        list(SERVICE_DEF_PROPERTIES)
                    )
                )
            service_name = key.replace("service.", "")
            services = options.get("service") or {}
            service_def_parts = [el for el in val.split('\n') if el]
            service_def = {}
            for part in service_def_parts:
                opt, value = part.split("=")
                opt = opt.strip()
                if opt not in SERVICE_DEF_PROPERTIES:
                    raise p_config.ConfigurationError(
                        "Unrecognized service property: '{}'".format(opt)
                    )
                value = value.strip()
                service_def[opt] = value
            defined_protocol = "protocol" in service_def
            defined_schema = "schema" in service_def
            if not (defined_protocol or defined_schema):
                raise p_config.ConfigurationError(
                    "Service must have either protocol or schema defined."
                )

            services[service_name] = service_def
            key = "service"
            val = services
        options[key] = val
    options["auto_compile"] = p_settings.asbool(options.get("auto_compile"))
    return options


def derive_service_path(service_name, url_pattern=None, path_prefix=None):
    """
    Given a service name, existing url_pattern, and url path prefix, derive a
    normalized/finalized url path based on them.

    :param service_name: an avro service name.
    :param url_pattern: an optional existing url pattern.
    :param path_prefix: an optional path prefix.
    :return: a normalized url path.
    """

    if not isinstance(service_name, basestring) or not service_name:
        raise ValueError("Service name must be a non-empty string.")

    path_parts = []
    if not isinstance(url_pattern, basestring) or not url_pattern:
        url_pattern = service_name

    if url_pattern.startswith("/"):
        url_pattern = url_pattern[1:]

    if isinstance(path_prefix, basestring) and path_prefix:
        if not path_prefix.startswith("/"):
            path_prefix = "/" + path_prefix
        path_parts.append(path_prefix)

    path_parts.append(url_pattern)

    url_pattern = "/".join(path_parts)
    if not url_pattern.startswith("/"):
        url_pattern = "/" + url_pattern

    return url_pattern


__all__ = [get_config_options.__name__]



import logging
import os
import shlex
import subprocess
import sys

from avro import protocol as avro_protocol

logger = logging.getLogger(__name__)


def run_subprocess_command(command, out_buffer=sys.stdout, exit_on_error=True):
    """
    Execute a subprocess based on the provided command.

    If an out_buffer is provided, we assume a buffer interface and write the
    contents of the subprocess's stdout to it.

    if exit_on_error is provided, we call sys.exit if the exit code of the
    executed subprocess was non-zero.

    This returns the exit code of the executed subprocess.

    :param command: a string or iterable with strings as a command to exec.
    :param out_buffer: an optional output buffer.
    :param exit_on_error: whether or not to kill this process on non-zero exit.
    :return: the exit code of the subprocess.
    """
    encoding = sys.stdout.encoding or "utf-8"
    if hasattr(command, "__iter__"):
        command = " ".join(command)

    command = shlex.split(command)
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT
    )

    while proc.poll() is None:
        line = proc.stdout.readline()[:-1].strip()
        if line and out_buffer:
            try:
                out_buffer.write(line.decode(encoding))
            except TypeError:
                out_buffer.write(line)
            out_buffer.write("\n")

    line = proc.stdout.read()[:-1].strip()
    if line and out_buffer:
        try:
            out_buffer.write(line.decode(encoding))
        except TypeError:
            out_buffer.write(line)
        out_buffer.write("\n")

    exit_code = proc.returncode
    if exit_on_error and exit_code != 0:
        sys.exit(exit_code)

    return exit_code


def compile_protocol(protocol, schema, jar_file):
    """
    Given the provided protocol path, schema path, and jar file path, attempt
    to compile the protocol file into an avro schema.

    :param protocol: an avro protocol file path.
    :param schema: a file path to write the compiled schema.
    :param jar_file: a jar file to use for the compilation.
    """

    if None in [protocol, schema, jar_file]:
        raise ValueError("Input must not be NoneType.")

    if not os.path.exists(jar_file):
        raise OSError("No such file or directory {}".format(jar_file))

    if not os.path.exists(protocol):
        raise OSError("No such file or directory {}".format(protocol))

    logger.debug("Compiling {} into {}".format(protocol, schema))
    command = [
        "java",
        "-jar",
        jar_file,
        "idl",
        protocol,
        schema
    ]
    run_subprocess_command(command)


def get_protocol_from_file(schema_path):
    """
    Given the provided schema path, load the schema contents into memory and
    parse them into an avro.protocol.Protocol object.

    :param schema_path: a schema file path.
    :return: an avro.protocol.Protocol object.
    """

    if not isinstance(schema_path, basestring):
        raise ValueError("Schema path must be of type {}".format(basestring))

    if not os.path.exists(schema_path):
        raise OSError("No such file or directory '{}'".format(schema_path))

    with open(schema_path) as _file:
        protocol = avro_protocol.parse(_file.read())

    return protocol


__all__ = [compile_protocol.__name__]



import inspect

from pyramid import view as p_view


class avro_message(p_view.view_config):
    """Decorator for registering an avro message with a given service."""

    _err = "service_name must be provided as an attribute or keyword " \
           "argument to the callable {}"

    def __call__(self, wrapped):
        """
        The wrapped object to register as a valid callable for handling this
        message.

        :param wrapped: an object to register as a message for a service.
        :return: the wrapped object.
        """
        settings = self.__dict__.copy()
        depth = settings.pop("_depth", 0)
        settings.pop("message_impl", None)

        # Mostly a copy of view_config"s impl, just changing the config
        # attachment.
        def callback(context, name, ob):
            message_impl = wrapped
            # Try to grab the "service_name" from our initial args if possible.
            if settings.get("service_name") is None:
                # If it's not in the args, see if it's on the object.
                service_name = getattr(ob, "service_name", None)

                # Try to use the class name if it's  class.
                if service_name is None:
                    if not inspect.isclass(ob):
                        raise AttributeError(
                            self._err.format(message_impl.__name__)
                        )
                    message_impl = message_impl.__get__(ob)
                    service_name = ob.__name__.lower()
                settings["service_name"] = service_name
            message_name = settings.pop("message", None)
            settings["message"] = message_name or message_impl.__name__
            config = context.config.with_package(info.module)
            config.register_avro_message(
                message_impl=message_impl,
                **settings
            )

        info = self.venusian.attach(
            wrapped,
            callback,
            category="pyramid",
            depth=depth + 1
        )
        return wrapped


__all__ = [avro_message.__name__]



import os

from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test as TestCommand


here = os.path.abspath(os.path.dirname(__file__))
test_dir = os.path.join(here, "tests")
__version__ = None


with open(os.path.join(here, "README.md")) as _file:
    README = _file.read().strip()
with open(os.path.join(here, "CHANGES.txt")) as _file:
    CHANGES = _file.read().strip()
with open(os.path.join(here, "requirements", "main.txt")) as _file:
    REQUIREMENTS = [l.strip() for l in _file.readlines()]
with open(os.path.join(here, "requirements", "test.txt")) as _file:
    TEST_REQUIREMENTS = REQUIREMENTS[:]
    TEST_REQUIREMENTS += [l.strip() for l in _file.readlines()]
with open(os.path.join(here, "pyramid_avro", "version.py")) as _file:
    exec(_file.read())


class PyTest(TestCommand):
    """Copied from py.test docs."""

    user_options = TestCommand.user_options + [
        ("with-coverage=", "c", "Generate coverage reports.")
    ]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.with_coverage = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_suite = self.test_suite or test_dir
        self.test_args = [self.test_suite]
        self.test_args += ["-rw"]

        if self.with_coverage:
            self.test_args += [
                "--cov={}".format("pyramid_avro"),
                "--cov-report={}".format(self.with_coverage)
            ]

    def run_tests(self):
        import pytest
        pytest.main(self.test_args)


setup(
    name="pyramid-avro",
    version=__version__,
    description="Avro RPC Bindings for Pyramid",
    long_description=README + "\n\n" + CHANGES,
    classifiers=[
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
    license="Apache License (2.0)",
    author="Alex Milstead",
    author_email="alex@amilstead.com",
    url="http://github.com/packagelib/pyramid-avro",
    keywords="pyramid avro",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    test_suite="tests",
    cmdclass={"test": PyTest}
)



import logging
import os
import traceback

from pyramid import config as p_config
from pyramid import exceptions as p_exc

from . import routes
from . import settings
from . import utils

logger = logging.getLogger(__name__)


def add_avro_route(config, service_name, pattern=None, protocol=None,
                   schema=None):
    """
    Queues an action to add an avro route to this config based on the provided
    parameters.

    These can come from either a configuration file or issued programmatically
    against the config after including pyramid_avro.

    This function performs a series of validations to ensure that the provided
    input is actually capable of being registered.

    General rules of thumb are:
        # You definitely need AT LEAST a schema file defined.
        # If you provide "auto_compile" in your INI file and:
            ## DID NOT provide tools_jar - BOOM
            ## DID NOT provide a protocol file - BOOM
        # Any file paths are subject to "No such file or directory" errors.

    :param config: a pyramid.config.Configurator object.
    :param service_name: a service name to register a route for.
    :param pattern: a url path pattern to register for this service.
    :param protocol: an optional Avro protocol file.
    :param schema: an optional Avro schema file.
    """

    avro_settings = settings.get_config_options(config.get_settings())
    service_def = avro_settings.get("service").get(service_name)
    no_service_def = protocol is None and schema is None
    if no_service_def and service_def is None:
        err = "Not enough information provided to register service '{}'. " \
              "Please provide an ini file definition or a " \
              "protocol/schema when calling 'add_avro_route'.".format(
                service_name)

        raise p_config.ConfigurationError(err)

    # Derive service url path.
    service_path = settings.derive_service_path(
        service_name,
        pattern,
        avro_settings["default_path_prefix"]
    )

    # Derive base directory for files.
    base_dir = avro_settings["protocol_dir"]
    if base_dir is None or not os.path.isabs(base_dir):
        parts = [os.path.dirname(config.root_package.__file__)]
        if base_dir is not None:
            parts.append(base_dir)
        base_dir = os.path.join(*parts)

    # Discover protocol and schema files.
    if protocol is not None:
        if not os.path.isabs(protocol):
            # Make sure this is an absolute path.
            protocol = os.path.join(base_dir, protocol)

        # If schema was none, "auto-discover" it.
        if schema is None:
            directory, filename = os.path.split(protocol)
            filename, ext = os.path.splitext(filename)
            filename = ".".join([filename, "avpr"])
            schema = os.path.join(directory, filename)

    # Normalize schema path.
    if schema is not None:
        if not os.path.isabs(schema):
            schema = os.path.join(base_dir, schema)

    auto_compile = avro_settings["auto_compile"]
    if not auto_compile and not os.path.exists(schema):
        raise p_config.ConfigurationError(
            "No such file or directory '{}'".format(schema)
        )

    if auto_compile:
        tools_jar = avro_settings["tools_jar"]
        if tools_jar is None:
            err = "Cannot auto_compile without tools_jar defined."
            raise p_config.ConfigurationError(err)

        if not os.path.exists(tools_jar):
            err = "No such file or directory: {}".format(tools_jar)
            raise p_config.ConfigurationError(err)

        if protocol is None:
            err = "Cannot auto_compile without a protocol defined."
            raise p_config.ConfigurationError(err)

    def register():
        # Begin route definition.
        route = ".".join(["avro", service_name])
        registry = config.registry
        # Shadow outer-scope
        protocol_file = protocol
        schema_file = schema

        if avro_settings["auto_compile"]:
            utils.compile_protocol(protocol_file, schema_file, tools_jar)

        try:
            with open(schema_file) as _file:
                schema_contents = _file.read()
        except Exception:
            message = traceback.format_exc()
            raise p_config.ConfigurationError(message)

        try:
            route_def = routes.AvroServiceRoute(route, schema_contents)
        except Exception:
            raise p_exc.ConfigurationError(
                "Failed to register route {}:\n {}".format(
                    service_name,
                    traceback.format_exc()
                )
            )

        registry.registerUtility(
            route_def,
            routes.IAvroServiceRoute,
            name=route
        )
        logger.debug("Registering avro service: {} => {}".format(
            route,
            service_path
        ))
        config.add_route(route, service_path, request_method="POST")
        config.add_view(route_name=route, view=route_def)

    config.action(
        ("avro-route", service_name),
        register,
        order=p_config.PHASE0_CONFIG
    )


def register_avro_message(config, service_name, message_impl, message=None):
    """
    Registers a callable object as the implementation for a message belonging
    to an avro protocol service whose route has been added with the above
    "add_avro_route" directive.

    The message implementation must be callable. It may be any object that is
    callable and accepts a request.

    If the message name is not explicitly provided, this method will attempt
    to pull it directly off of the "__name__" attribute of the provided
    "message_impl".

    :param config: a pyramid.config.Configurator object.
    :param service_name: an avro service name added with add_avro_route.
    :param message_impl: an implementation for message.
    :param message: an optional message name.
    :return:
    """

    if isinstance(message_impl, str):
        message_impl = config.maybe_dotted(message_impl)

    if not callable(message_impl):
        raise p_exc.ConfigurationError(
            "{} must be callable.".format(message_impl)
        )

    message_name = message or message_impl.__name__
    route = ".".join(["avro", service_name])

    def register():
        registry = config.registry
        route_def = registry.queryUtility(routes.IAvroServiceRoute, name=route)
        if route_def is None:
            err = "Service '{}' has no route defined.".format(service_name)
            raise p_exc.ConfigurationError(err)

        logger.debug("Registering message {} for service {}".format(
            message_name,
            route
        ))
        route_def.register_message_impl(message_name, message_impl)

    config.action(
        ("avro-message", service_name, message_name),
        register,
        order=p_config.PHASE0_CONFIG
    )


def includeme(config):
    """
    Adds directives:
        # add_avro_route
        # register_avro_message

    Scans the provided settings for any pre-defined services and adds them at
    this step.

    :param config: a pyramid.config.Configurator object.
    """
    config.add_directive("add_avro_route", add_avro_route)
    config.add_directive("register_avro_message", register_avro_message)
    options = settings.get_config_options(config.get_settings())
    service_defs = options.get("service") or {}
    for service_name, service_opts in service_defs.items():
        config.add_avro_route(service_name, **service_opts)

    logger.debug("Finished preparing for avro services.")


__all__ = [includeme.__name__]
